//----------------------------------------------------------------------------
//  Project Dockcbex
//  Tenermerx Software
//  Copyright  1997 Tenermerx Software.
//
//  SUBSYSTEM:    Dockcbex Application
//  FILE:         dcbapp.cpp
//  AUTHOR:       Christopher Kohlhoff
//
//  OVERVIEW
//  ~~~~~~~~
//  Source file for implementation of TDCBApp (TApplication).
//
//----------------------------------------------------------------------------

#include <owl/pch.h>

#include <owl/buttonga.h>
#include <owl/decframe.h>
#include <owl/editfile.h>
#include "dcbapp.h"


const int kToolbarId = 1234;


// TMyToolbar class
// ~~~~~~~~~~~~~~~~
// A dockable toolbar close that remembers its position
// when it is hidden/shown
//
class TMyToolbar : public TDockableControlBar
{
public:
  TMyToolbar(TWindow* parent) : TDockableControlBar(parent) {}

  void Hide()
  {
    GetHarbor()->Remove(*this);
  }

  void Show(THarbor* harbor)
  {
    harbor->Insert(*this, Location, &Rect.TopLeft());
    harbor->Move(*this, Location, &Rect.TopLeft());
  }

  void SavePosition()
  {
    // get the absolute location of the toolbar
    //
    TDockingSlip* dockingSlip = TYPESAFE_DOWNCAST(Parent, TDockingSlip);
    Location = (dockingSlip ? dockingSlip->GetLocation() : alTop);

    // get the rectangle of the toolbar
    //
    TFloatingSlip* floatingSlip = TYPESAFE_DOWNCAST(Parent, TFloatingSlip);
    if (floatingSlip) floatingSlip->GetWindowRect(Rect);
    else GetRect(Rect);
  }

  TResult WindowProc(uint msg, TParam1 p1, TParam2 p2)
  {
    if (msg == WM_SHOWWINDOW && !p1) // we are being hidden
      SavePosition();
    return TDockableControlBar::WindowProc(msg, p1, p2);
  }

protected:
  TRect Rect;
  TAbsLocation Location;
};




//{{TDCBApp Implementation}}


//
// Build a response table for all messages/commands handled
// by the application.
//
DEFINE_RESPONSE_TABLE1(TDCBApp, TApplication)
//{{TDCBAppRSP_TBL_BEGIN}}
  EV_COMMAND(CM_FILENEW, CmFileNew),
  EV_COMMAND(CM_FILEOPEN, CmFileOpen),
  EV_COMMAND(CM_VIEWTOOLBAR, CmViewToolbar),
  EV_COMMAND_ENABLE(CM_VIEWTOOLBAR, CeViewToolbar),
//{{TDCBAppRSP_TBL_END}}
END_RESPONSE_TABLE;



//--------------------------------------------------------
// TDCBApp
//
TDCBApp::TDCBApp() : TApplication("TDockableControlBar Example")
{

  // Common file file flags and filters for Open/Save As dialogs.  Filename and directory are
  // computed in the member functions CmFileOpen, and CmFileSaveAs.
  //
  FileData.Flags = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT;
  FileData.SetFilter(string(GetInstance(), IDS_FILEFILTER).c_str());
  FileData.DefExt = "txt";

#ifdef BI_PLAT_WIN32
  FileData.Flags |= OFN_EXPLORER;
#endif
}


TDCBApp::~TDCBApp()
{
}


void TDCBApp::SetupSpeedBar(TDecoratedFrame* frame)
{
  ApxHarbor = new THarbor(*frame);

  // Create default toolbar New and associate toolbar buttons with commands.
  //
  Toolbar = new TMyToolbar(frame);
  Toolbar->Insert(*new TButtonGadget(CM_FILENEW, CM_FILENEW));
  Toolbar->Insert(*new TButtonGadget(CM_FILEOPEN, CM_FILEOPEN));
  Toolbar->Insert(*new TButtonGadget(CM_FILESAVE, CM_FILESAVE));
  Toolbar->Insert(*new TSeparatorGadget(6));
  Toolbar->Insert(*new TButtonGadget(CM_EDITCUT, CM_EDITCUT));
  Toolbar->Insert(*new TButtonGadget(CM_EDITCOPY, CM_EDITCOPY));
  Toolbar->Insert(*new TButtonGadget(CM_EDITPASTE, CM_EDITPASTE));
  Toolbar->Insert(*new TSeparatorGadget(6));
  Toolbar->Insert(*new TButtonGadget(CM_EDITUNDO, CM_EDITUNDO));
  Toolbar->Insert(*new TSeparatorGadget(6));
  Toolbar->Insert(*new TButtonGadget(CM_EDITFIND, CM_EDITFIND));
  Toolbar->Insert(*new TButtonGadget(CM_EDITREPLACE, CM_EDITREPLACE));

  // Add caption and fly-over help hints.
  //
  Toolbar->SetCaption("Toolbar");
  Toolbar->SetHintMode(TGadgetWindow::EnterHints);

  // Resize the toolbar so that it floats horizontally
  //
  TRect rect = Toolbar->GetWindowRect();
  TSize size(rect.Width(), rect.Height());
  Toolbar->Layout(alTop, &size);

  ApxHarbor->Insert(*Toolbar, alTop);
}


//--------------------------------------------------------
// TDCBApp
// ~~~~~~~
// Application intialization.
//
void TDCBApp::InitMainWindow()
{
  if (nCmdShow != SW_HIDE)
    nCmdShow = (nCmdShow != SW_SHOWMINNOACTIVE) ? SW_SHOWNORMAL : nCmdShow;

  // Create the client and frame window
  //
  TEditFile* client = new TEditFile(0, 0, 0);
  client->Attr.ExStyle = WS_EX_CLIENTEDGE;
  TDecoratedFrame* frame = new TDecoratedFrame(0, GetName(), client, false);

  // Assign icons for this application.
  //
  frame->SetIcon(this, IDI_SDIAPPLICATION);
  frame->SetIconSm(this, IDI_SDIAPPLICATION);

  // Menu associated with window and accelerator table associated with table.
  //
  frame->AssignMenu(IDM_SDI);

  // Associate with the accelerator table.
  //
  frame->Attr.AccelTable = IDM_SDI;

  SetupSpeedBar(frame);

  SetMainWindow(frame);

  frame->SetMenuDescr(TMenuDescr(IDM_SDI));
}



//--------------------------------------------------------
// TDCBApp
// ~~~~~~~
// Menu File New command
//
void TDCBApp::CmFileNew()
{
  TEditFile* client = TYPESAFE_DOWNCAST(GetMainWindow()->GetClientWindow(), TEditFile);
  client->NewFile();
}


//--------------------------------------------------------
// TDCBApp
// ~~~~~~~
// Menu File Open command
//
void TDCBApp::CmFileOpen()
{
  // Display standard Open dialog box to select a file name.
  //
  *FileData.FileName = 0;

  TEditFile* client = TYPESAFE_DOWNCAST(GetMainWindow()->GetClientWindow(), TEditFile);
  if (client->CanClose())
    if (TFileOpenDialog(GetMainWindow(), FileData).Execute() == IDOK)
      client->ReplaceWith(FileData.FileName);
}


//--------------------------------------------------------
// TDCBApp
// ~~~~~~~
// Hide or show toolbar menu command
//
void TDCBApp::CmViewToolbar()
{
  if (Toolbar->IsWindowVisible()) Toolbar->Hide();
  else Toolbar->Show(ApxHarbor);
}

void TDCBApp::CeViewToolbar(TCommandEnabler &tce)
{
  tce.SetCheck(Toolbar->IsWindowVisible());
}



int OwlMain(int , char* [])
{
  TDCBApp app;
  return app.Run();
}

