/***********************************************************************************
 *		  																								     *
 *	File:			SB.cpp.			  					 												  *
 *																											  *
 * Contents: 	Implementation of TRange, TSpinEdit TSpinScrollBar and TSpinButton. *
 *																											  *
 * Purpose:  	Be able to spin with doubles.													  *
 *                                                                                 *
 *	Features: 	Dimension (like 122.5 h, 136.9 $ and 45.23 g/sqrm).                 *
 *					Separator (like 254,34, 23|23, 67*23, 78'32).                       *
 *             Fixed width (like 23,500, 456,10, 523,00, 73,000).                  *
 *             Extensive error correction (like 1  3,,,5 -> 13,5, ...6,,9 -> 6,9). *
 *             Popup menu (Top|Middle|Bottom|Current menu above arrows).			  *
 *					Error dialog when nonsense are written.                             *
 *																											  *
 *					By default only the digits 0,..., 9, a sign and a dimension are     *
 *					allowed to go in the TEdit. Most things are optional from the		  *
 *					construction of the TSpinButton.												  *
 *																											  *
 * Remarks:    When you look at parameters of constructors Ex'ed data are used for *
 *					*this and the rest are given away to lower object(s).     			  *
 *																											  *
 * Originator: Michael Mogensen, BassetSoft Denmark, Hilleroed, Denmark.			  *
 *																											  *
 *	E-mail:		mogensen@post6.tele.dk.	      												  *
 *																											  *
 * Compiler:	Borland C++ 5.01.																	  *
 *																											  *
 * Version:		2.0.																	 				  *
 *																											  *
 * Period:		98.07.10 - 98.07.13, 98.08.13 - 98.08.17. 								  *
 *																											  *
 ***********************************************************************************/

/***********************************************************************************
 *	Miscellaneous.													 									  *
 ***********************************************************************************/

/***********************************************************************************
 *	Headers.															 									  *
 ***********************************************************************************/
#pragma hdrstop
#include "SB.h"

/***********************************************************************************
 *	TRange.														 										  *
 ***********************************************************************************/

// 1. Object. (alphabetical).
TRange::TRange(const double dBottomEx,
       			const double dCurrentEx,
       			const double dTopEx,
       			const double dStepEx,
               const int iSignificantEx) :
		 			dBottom(dBottomEx),
     	 			dCurrent(dCurrentEx),
          		dTop(dTopEx),
  		 	 		dStep(dStepEx),
               iSignificant(iSignificantEx),
               bTouchZero(false)
   // Create.
	{ Normalize(); }

TRange::~TRange()
// Delete.
{}

// 4. Slaves. (alphabetical).
const void TRange::Normalize()
{
	if(dBottom > dTop)
		SwapObjects(dBottom, dTop);
   SupressObjectInRange(dCurrent, dBottom, dTop);
   dStep = Abs(dStep);
   dStep = max(I_DStepMin, dStep);
   if(dStep > (dTop - dBottom) / 2.0)
   	dStep = I_DStep;
}

/***********************************************************************************
 *	TSpinEdit.														 									  *
 ***********************************************************************************/

DEFINE_RESPONSE_TABLE1(TSpinEdit, TEdit)
	EV_WM_KILLFOCUS,
	EV_WM_SETFOCUS,
END_RESPONSE_TABLE;

// 1. Object. (alphabetical).
TSpinEdit::TSpinEdit(TWindow *pParent,
      				   const int iId,
		   				const double dBottom,
      		    		const double dCurrent,
          				const double dTop,
		          		const double dStep,
                     const bool bErrorBoxEx,
                     const int iPrecisionEx,
                     const string sDimensionEx,
                     const string sSeparatorEx,
      		    		const int iSignificant,
               		const string sEditEx) :
                     bErrorBox(bErrorBoxEx),
                     iPrecision(iPrecisionEx),
                     sDimension(sDimensionEx),
                     sSeparator(sSeparatorEx),
						 	sSeparatorNorm("."),
						 	sSpace(" "),
						   uError(IDNO),
                     TEdit(pParent, iId - 1)
   // Create.
	{
      // Set range.
		pRange = new TRange(dBottom,
      						  dCurrent,
                          dTop,
                          dStep,
                          iSignificant);
      // Set full validator string.
      string sValidator(sSpace + sDimension + sSeparator + sEditEx);
      if(!sValidator.is_null())
      	SetValidator(new TFilterValidator(sValidator.c_str()));
      // First validate when focus are lost.
      TValidator *pValidator = GetValidator();
      if(pValidator && pValidator->HasOption(voOnAppend))
      	pValidator->UnsetOption(voOnAppend);
   }

TSpinEdit::~TSpinEdit()
// Delete.
{
	if(pRange)
   	delete pRange;
}

// 2. Event's. (alphabetical).
void TSpinEdit::EvKillFocus(HWND hWndGetFocus)
// Lost the focus.
{
	TEdit::EvKillFocus(hWndGetFocus);
   // Update it.
	SpinTo(pRange->dCurrent, true, false);
}

void TSpinEdit::EvSetFocus(HWND hWndLostFocus)
// Got the focus.
{
	TEdit::EvSetFocus(hWndLostFocus);
}

bool TSpinEdit::IdleAction(long lIdleCount)
// System idle.
{
   if(uError == IDYES)
   {
	   char pszBuffer[I_StringLen + 1];
   	strcpy(pszBuffer, "");
		string sBottom(gcvt(pRange->dBottom, pRange->iSignificant, pszBuffer));
		strcpy(pszBuffer, "");
		string sTop(gcvt(pRange->dTop, pRange->iSignificant, pszBuffer));
		strcpy(pszBuffer, "");
		wsprintf(pszBuffer, I_SError_Text, (LPSTR)sBottom.c_str(), (LPSTR)sTop.c_str(), (LPSTR)sDimension.c_str());
		::MessageBox(HWindow, pszBuffer, I_SError_Caption, MB_ICONEXCLAMATION);
	   SetText(pRange->sCurrent.c_str());
     	uError = IDNO;
   }
	return TEdit::IdleAction(lIdleCount);
}

void TSpinEdit::SetupWindow()
// Setup.
{
	TEdit::SetupWindow();
   // Update it.
   SpinTo(pRange->dCurrent, false);
}

// 4. Slaves. (alphabetical).
const void TSpinEdit::SpinTo(double dCurrentNew, const bool bRead, const bool bUpDown)
// Up/down.
{
   static char pszBuffer[I_StringLen + 1];
	if(bRead && IsModified())
   {
      // Go from decorated format into normal format.
	   strcpy(pszBuffer, "");
		TEdit::GetText(pszBuffer, I_StringLen);
   	string sBufferClean(""),
      		 sBufferDirty(pszBuffer);
      size_t uId,
      		 uIdEnd;
      // Only accept one sign.
      string sSign("-");
  	  	if(sBufferDirty.contains(sSign))
      	sBufferClean += sSign;
     	// Only accept digit or one separator.
      string sDigit("0123456789");
      uIdEnd = sBufferDirty.length();
      for(uId = 0; uId < uIdEnd; uId++)
     		if(sDigit.contains(sBufferDirty[uId]))
      		sBufferClean += sBufferDirty[uId];
         else
         	if(sSeparator.contains(sBufferDirty[uId]) && !sBufferClean.contains(sSeparatorNorm))
	      		sBufferClean += sSeparatorNorm;
      // Check.
      uIdEnd = sBufferClean.length();
      for(uId = 0;
      	 uId < uIdEnd && !sDigit.contains(sBufferClean[uId]);
          uId++) {}
   	if(uId < uIdEnd)
   	{
	      // It was clean.
	      char *pszHelper;
         double dCurrent = strtod(sBufferClean.c_str(), &pszHelper),
         		 dStep = bUpDown ? pRange->dStep : 0.0;
	      dCurrentNew = dCurrentNew > pRange->dCurrent ? dCurrent + dStep :
         															  dCurrent - dStep;
		}
      else
      	// It was dirty.
      	dCurrentNew = HUGE_VAL;
      if(dCurrentNew >= HUGE_VAL || dCurrentNew <= -HUGE_VAL)
      {
      	if(bErrorBox)
         	uError = IDYES;
         else
         	uError = IDNO;
         return;
      }
   }
   // New current.
	pRange->dCurrent = dCurrentNew;
	SupressObjectInRange(pRange->dCurrent, pRange->dBottom, pRange->dTop);
   if(pRange->bTouchZero && Abs(pRange->dCurrent) < pRange->dStep || Abs(pRange->dCurrent) < I_DStepMin)
   	pRange->dCurrent = 0.0;
   // Go from normal format into decorated format.
   strcpy(pszBuffer, "");
   // Insert user width if wanted.
   if(iPrecision > 0)
   	sprintf(pszBuffer, "%.*f", iPrecision, pRange->dCurrent);
   else
	   gcvt(pRange->dCurrent, pRange->iSignificant, pszBuffer);
   string sBuffer(pszBuffer);
   pRange->sCurrent = sBuffer;
   // Insert space and dimension, if wanted.
   if(!sDimension.is_null())
   {
      pRange->sCurrent += sSpace;
      pRange->sCurrent += sDimension;
   }
   // Insert separator, if wanted (if not a normal one).
	if(!sSeparator.is_null() && sSeparator != sSeparatorNorm)
   {
		size_t uId = pRange->sCurrent.find(sSeparatorNorm);
      if(uId != NPOS)
      	pRange->sCurrent.replace(uId, 1, sSeparator);
   }
   // It's decorated. Show it.
   SetText(pRange->sCurrent.c_str());
}

/***********************************************************************************
 *	TSpinScrollBar.												 									  *
 ***********************************************************************************/

DEFINE_RESPONSE_TABLE1(TSpinScrollBar, TScrollBar)
	EV_WM_HSCROLL,
	EV_WM_KILLFOCUS,
	EV_WM_RBUTTONDOWN,
	EV_WM_VSCROLL,
	EV_COMMAND(I_Menu_Top, Menu_Top),
	EV_COMMAND_ENABLE(I_Menu_Top, Menu_Top_Enable),
	EV_COMMAND(I_Menu_Middle, Menu_Middle),
	EV_COMMAND_ENABLE(I_Menu_Middle, Menu_Middle_Enable),
	EV_COMMAND(I_Menu_Bottom, Menu_Bottom),
	EV_COMMAND_ENABLE(I_Menu_Bottom, Menu_Bottom_Enable),
	EV_COMMAND(I_Menu_Current, Menu_Current),
END_RESPONSE_TABLE;

// 1. Object. (alphabetical).
TSpinScrollBar::TSpinScrollBar(TWindow *pParent,
										 const int iId,
				   					 const double dBottom,
          							 const double dCurrent,
				         			 const double dTop,
          							 const double dStep,
                     			 const bool bErrorBox,
              						 const bool bMenuEx,
                     			 const int iPrecision,
                     			 const string sDimension,
                     			 const string sSeparator,
      		    					 const int iSignificant,
               					 const string sEdit) :
                     			 TScrollBar(pParent, iId + 1)
   // Create.
	{
   	pSpinEdit = new TSpinEdit(pParent,
      								  iId,
                                dBottom,
                                dCurrent,
                                dTop,
                                dStep,
                                bErrorBox,
                                iPrecision,
                                sDimension,
                                sSeparator,
                                iSignificant,
                                sEdit);
      if(bMenuEx)
	   	pMenu = new TPopupMenu(*new TMenu(GetApplication()->GetInstance(), (TResId)I_Menu_Scroller));
		else
			pMenu = NULL;
   }

TSpinScrollBar::~TSpinScrollBar()
// Delete.
{
	if(pSpinEdit)
   	delete pSpinEdit;
   if(pMenu)
   	delete pMenu;
}

// 2. Event's. (alphabetical).
void TSpinScrollBar::EvHScroll(UINT uScrollCode, UINT uThumbPosition, HWND hWindowCtl)
// Respond to left/right events (for special spinner look).
{
	TScrollBar::EvHScroll(uScrollCode, uThumbPosition, hWindowCtl);
	switch(uScrollCode)
	{
		case SB_LINEUP: (*pSpinEdit)--; return;
		case SB_LINEDOWN: (*pSpinEdit)++; return;
	}
}

void TSpinScrollBar::EvRButtonDown(UINT uKey, TPoint &ipMouse)
// Right button down in client area (above scrollbar).
{
   TScrollBar::EvRButtonDown(uKey, ipMouse);
	// Show menu.
   ClientToScreen(ipMouse);
   if(pMenu)
   	pMenu->TrackPopupMenu(TPM_LEFTALIGN | TPM_RIGHTBUTTON, ipMouse, 0, HWindow);
}

void TSpinScrollBar::EvVScroll(UINT uScrollCode, UINT uThumbPosition, HWND hWindowCtl)
// Respond to up/down events (for normal spinner look).
{
	TScrollBar::EvVScroll(uScrollCode, uThumbPosition, hWindowCtl);
	switch(uScrollCode)
	{
		case SB_LINEUP: (*pSpinEdit)++; return;
		case SB_LINEDOWN: (*pSpinEdit)--; return;
	}
}

void TSpinScrollBar::SetupWindow()
// Setup.
{
	TScrollBar::SetupWindow();
}

/***********************************************************************************
 *	TSpinButton.													 									  *
 ***********************************************************************************/

// 1. Object. (alphabetical).
TSpinButton::TSpinButton(TWindow *pParent,
								 const int iId,
                  		 const double dBottom,
								 const double dCurrent,
                  		 const double dTop,
                   		 const double dStep,
                  		 const bool bErrorBox,
        						 const bool bMenu,
                  		 const int iPrecision,
                  		 const string sDimension,
                  		 const string sSeparator,
                  		 const int iSignificant,
                  		 const string sEdit)
	// Create.
	{
   	pSpinScrollBar = new TSpinScrollBar(pParent,
      												iId,
                                          dBottom,
                                          dCurrent,
                                          dTop,
                                          dStep,
                                          bErrorBox,
                                    		bMenu,
                                          iPrecision,
                                          sDimension,
                                          sSeparator,
                                          iSignificant,
                                          sEdit);
   }

TSpinButton::~TSpinButton()
// Delete.
{
	if(pSpinScrollBar)
   	delete pSpinScrollBar;
}

/***********************************************************************************
 *																		 									  *
 ***********************************************************************************/

// 0. Data. (alphabetical).
// 1. Object. (alphabetical).
// 2. Event's. (alphabetical).
// 3.0. Menu choice. (menu order).
// 3.1. Menu choice, enablers. (menu order).
// 4. Slaves. (alphabetical).
// 5. Other. (alphabetical).

