/***********************************************************************************
 *		  																								     *
 *	File:			SB.h.			  				 														  *
 *																											  *
 * Contents: 	Interface of TRange, TSpinEdit TSpinScrollBar and TSpinButton. 	  *
 *																											  *
 * Purpose:  	Be able to spin with doubles.													  *
 *																											  *
 *	Features: 	Dimension (like 122.5 h, 136.9 $ and 45.23 g/sqrm).                 *
 *					Separator (like 254,34, 23|23, 67*23, 78'32).                       *
 *             Fixed width (like 23,500, 456,10, 523,00, 73,000).                  *
 *             Extensive error correction (like 1  3,,,5 -> 13,5, ...6,,9 -> 6,9). *
 *             Popup menu (Top|Middle|Bottom|Current menu above arrows).			  *
 *					Error dialog when nonsense are written.                             *
 *																											  *
 *					By default only the digits 0,..., 9, a sign and a dimension are     *
 *					allowed to go in the TEdit. Most things are optional from the		  *
 *					construction of the TSpinButton.												  *
 *																											  *
 * Remarks:    -												 										  *
 *																											  *
 * Originator: Michael Mogensen, BassetSoft Denmark, Hilleroed, Denmark.			  *
 *																											  *
 *	E-mail:		mogensen@post6.tele.dk.	      												  *
 *																											  *
 * Version:		2.0.																	 				  *
 *																											  *
 * Period:		98.07.10 - 98.07.13, 98.08.13 - 98.08.17. 							 	  *
 *																											  *
 ***********************************************************************************/

/***********************************************************************************
 *	Miscellaneous.													 									  *
 ***********************************************************************************/
#define STRICT

#ifndef SB_h
#define SB_h

/***********************************************************************************
 *	Headers, OWL.													 									  *
 ***********************************************************************************/
#include <owl\edit.h>     				// TEdit.
#include <owl\menu.h>               // TPopupMenu.
#include <owl\scrollba.h>     		// TScrollBar.
#include <owl\validate.h>     		// TFilterValidator.

/***********************************************************************************
 *	Headers, C.													 										  *
 ***********************************************************************************/
#include <cstring.h>               	// string.
#include <errno.h>                  // HUGE_VAL
#include <math.h>                   // HUGE_VAL.
#include <stdio.h>                  // sprintf(...).
#include <stdlib.h>						// max(...), min(...) etc.

/***********************************************************************************
 *	Private.																 								  *
 ***********************************************************************************/
#include "generic.h"     				// Templates.
#include "sb.rh"                    // Identifiers.

/***********************************************************************************
 *	TRange.															 								 	  *
 ***********************************************************************************/
struct TRange
{
	// 0. Data. (alphabetical).
	bool bTouchZero;
	double dBottom,
     		 dCurrent,
          dTop,
  		 	 dStep;
   const int iSignificant;
   string sCurrent;
	// 1. Object. (alphabetical).
   TRange(const double,
          const double,
          const double,
          const double,
          const int);
   ~TRange();
	// 4. Slaves. (alphabetical).
	const void Normalize();
};

/***********************************************************************************
 *	TSpinEdit.															 								  *
 ***********************************************************************************/
class TSpinEdit : public TEdit
{
	protected:
		// 0. Data. (alphabetical).
      const bool bErrorBox;
      const int iPrecision;
      const string sDimension,
       		 		 sSeparator,
						 sSeparatorNorm,
						 sSpace;
   	UINT uError;
		// 2. Event's. (alphabetical).
		virtual void EvKillFocus(HWND);
		virtual void EvSetFocus(HWND);
		virtual bool IdleAction(long);
		virtual void SetupWindow();
	public:
		// 0. Data. (alphabetical).
      TRange *pRange;
		// 1. Object. (alphabetical).
		TSpinEdit(TWindow*,
					 const int,
                const double,
                const double,
                const double,
                const double,
                const bool,
                const int,
                const string,
                const string,
                const int,
                const string);
      virtual ~TSpinEdit();
		// 4. Slaves. (alphabetical).
		inline const TSpinEdit& operator++(int)
		{ SpinTo(pRange->dCurrent + pRange->dStep); return *this; }
		inline const TSpinEdit& operator--(int)
		{ SpinTo(pRange->dCurrent - pRange->dStep); return *this; }
      inline const HWND GetHandle()
      { return TWindow::GetHandle(); }
      inline const void SetHandle(HWND hWnd)
      { TWindow::SetHandle(hWnd); }
		const void SpinTo(double,
      						const bool bReadEx = true,
                        const bool bUpDownEx = true);
	// 5. Other. (alphabetical).
   DECLARE_RESPONSE_TABLE(TSpinEdit);
};

/***********************************************************************************
 *	TSpinScrollBar.														 							  *
 ***********************************************************************************/
class TSpinScrollBar : public TScrollBar
{
	protected:
		// 0. Data. (alphabetical).
      TPopupMenu *pMenu;
		// 2. Event's. (alphabetical).
      virtual void EvHScroll(UINT, UINT, HWND);
		virtual void EvRButtonDown(UINT, TPoint&);
      virtual void EvVScroll(UINT, UINT, HWND);
		virtual void SetupWindow();
		// 3.0. Menu choice. (menu order).
		inline void Menu_Top()
		{ pSpinEdit->SpinTo(pSpinEdit->pRange->dTop, false); }
		inline void Menu_Middle()
		{
		   double dMiddle = 0.5 * (pSpinEdit->pRange->dTop - pSpinEdit->pRange->dBottom);
			pSpinEdit->SpinTo(dMiddle, false);
		}
		inline void Menu_Bottom()
		{ pSpinEdit->SpinTo(pSpinEdit->pRange->dBottom, false); }
		inline void Menu_Current()
		{ pSpinEdit->SpinTo(pSpinEdit->pRange->dCurrent, false); }
		// 3.1. Menu choice, enablers. (menu order).
		inline void Menu_Top_Enable(TCommandEnabler &Command)
      { Command.Enable(pSpinEdit->pRange->dCurrent != pSpinEdit->pRange->dTop); }
		inline void Menu_Middle_Enable(TCommandEnabler &Command)
      {
		   double dMiddle = 0.5 * (pSpinEdit->pRange->dTop - pSpinEdit->pRange->dBottom);
      	Command.Enable(pSpinEdit->pRange->dCurrent != dMiddle);
      }
		inline void Menu_Bottom_Enable(TCommandEnabler &Command)
      { Command.Enable(pSpinEdit->pRange->dCurrent != pSpinEdit->pRange->dBottom); }
   public:
		// 0. Data. (alphabetical).
      TSpinEdit *pSpinEdit;
		// 1. Object. (alphabetical).
   	TSpinScrollBar(TWindow*,
							const int,
				   		const double,
          				const double,
				         const double,
          				const double,
                     const bool,
                     const bool,
                     const int,
                     const string,
                     const string,
                     const int,
                     const string);
      virtual ~TSpinScrollBar();
      inline const void SetHandle(HWND hWnd)
      { TWindow::SetHandle(hWnd); }
	// 5. Other. (alphabetical).
   DECLARE_RESPONSE_TABLE(TSpinScrollBar);
};

/***********************************************************************************
 *	TSpinButton.														 								  *
 ***********************************************************************************/
class TSpinButton
{
	protected:
		// 0. Data. (alphabetical).
      TSpinScrollBar *pSpinScrollBar;
   public:
		// 1. Object. (alphabetical).
   	TSpinButton(TWindow*, // Parent pointer.
						const int, // Control id.
				   	const double dBottomEx = I_DBottom, // From this.
          			const double dCurrentEx = I_DCurrent, // Begin at.
				      const double dTopEx = I_DTop, // To this.
          			const double dStepEx = I_DStep, // Increase with.
                  const bool bErrorBoxEx = true, // Show dialog box on invalid input or not.
                  const bool bMenuEx = true, // Show menu above scroller or not.
                  const int iPrecisionEx = I_IPrecision, // Number of digits after separator.
                  const string sDimensionEx = I_SDimension, // Unit.
                  const string sSeparatorEx = I_SSeparator, // Separator to use.
                  const int iSignificantEx = I_ISignificant, // When converting floating point number to a string.
                  const string sEditEx = I_SEdit); // Valid input for edit.
		~TSpinButton();
		// 4. Slaves. (alphabetical).
		inline const void Enable(const bool bState)
      {
      	pSpinScrollBar->EnableWindow(bState);
      	pSpinScrollBar->pSpinEdit->EnableWindow(bState);
      }
		inline const double& GetBottom()
      { return pSpinScrollBar->pSpinEdit->pRange->dBottom; }
		inline const double& GetCurrent()
      { return pSpinScrollBar->pSpinEdit->pRange->dCurrent; }
      inline const HWND GetHandleEdit()
      { return pSpinScrollBar->pSpinEdit->GetHandle(); }
      inline const HWND GetHandleScroller()
      { return pSpinScrollBar->GetHandle(); }
		inline const double GetMiddle()
      {
      	return(0.5 * (pSpinScrollBar->pSpinEdit->pRange->dTop -
      					  pSpinScrollBar->pSpinEdit->pRange->dBottom));
      }
		inline const double GetRange()
      { return(pSpinScrollBar->pSpinEdit->pRange->dTop - pSpinScrollBar->pSpinEdit->pRange->dBottom); }
		inline const void GetRange(double &dBottomEx, double &dTopEx)
      {
      	dBottomEx = pSpinScrollBar->pSpinEdit->pRange->dBottom;
         dTopEx = pSpinScrollBar->pSpinEdit->pRange->dTop;
      }
		inline const double& GetTop()
      { return pSpinScrollBar->pSpinEdit->pRange->dTop; }
		inline const double& GetStep()
      { return pSpinScrollBar->pSpinEdit->pRange->dStep; }
		inline const char* GetText()
      { return pSpinScrollBar->pSpinEdit->pRange->sCurrent.c_str(); }
		inline const void SetBottom(const double &dBottomEx)
      {
      	pSpinScrollBar->pSpinEdit->pRange->dBottom = dBottomEx;
      	pSpinScrollBar->pSpinEdit->pRange->Normalize();
      }
		inline const void SetCurrent(const double &dCurrentEx)
      { pSpinScrollBar->pSpinEdit->SpinTo(dCurrentEx); }
      inline const void SetHandleEdit(HWND hWnd)
      { pSpinScrollBar->pSpinEdit->SetHandle(hWnd); }
      inline const void SetHandleScroller(HWND hWnd)
      { pSpinScrollBar->SetHandle(hWnd); }
		inline const void SetRange(double &dBottomEx, double &dTopEx)
      {
      	pSpinScrollBar->pSpinEdit->pRange->dBottom = dBottomEx;
         pSpinScrollBar->pSpinEdit->pRange->dTop = dTopEx;
      	pSpinScrollBar->pSpinEdit->pRange->Normalize();
      }
		inline const void SetTop(const double &dTopEx)
      {
      	pSpinScrollBar->pSpinEdit->pRange->dTop = dTopEx;
      	pSpinScrollBar->pSpinEdit->pRange->Normalize();
      }
		inline const void SetStep(double &dStepEx)
      {
      	pSpinScrollBar->pSpinEdit->pRange->dStep = dStepEx;
      	pSpinScrollBar->pSpinEdit->pRange->Normalize();
      }
	   inline const void TouchZero(const bool bTouchZeroEx)
   	{ pSpinScrollBar->pSpinEdit->pRange->bTouchZero = bTouchZeroEx; }
      inline const void Update()
      { pSpinScrollBar->pSpinEdit->SetText(GetText()); }
};

/***********************************************************************************
 *			 															 									  *
 ***********************************************************************************/


// 0. Data. (alphabetical).
// 1. Object. (alphabetical).
// 2. Event's. (alphabetical).
// 3.0. Menu choice. (menu order).
// 3.1. Menu choice, enablers. (menu order).
// 4. Slaves. (alphabetical).
// 5. Other. (alphabetical).

#endif

