//----------------------------------------------------------------------------
//  Project Shitemex
//  Tenermerx Software
//  Copyright  1997 Tenermerx Software.
//
//  SUBSYSTEM:    Shitemex Application
//  FILE:         shexwndw.cpp
//  AUTHOR:       Christopher Kohlhoff
//
//  OVERVIEW
//  ~~~~~~~~
//  Source file for implementation of TShExWindow (TWindow).
//
//----------------------------------------------------------------------------

#include <owl/pch.h>
#include <owl/shellitm.h>
#include <owl/inputdia.h>
#include <dir.h>

#include "shexapp.h"
#include "shexwndw.h"


const int kIconXOffset = 10;
const int kIconYOffset = 10;
const int kTextXOffset = 50;
const int kTextYOffset = 10;
const int kTextHeight  = 20;

// context menu identifiers
//
const int kContextBase = 28000;
const int kContextMax  = 28100;

//
// Build a response table for all messages/commands handled by the application.
//
DEFINE_RESPONSE_TABLE1(TShExWindow, TWindow)
//{{TShExWindowRSP_TBL_BEGIN}}
  EV_COMMAND(CM_FILEOPEN, CmFileOpen),
  EV_COMMAND(CM_FILEOPENFOLDER, CmFileOpenFolder),
  EV_COMMAND(CM_SHADDTODOCS, CmAddToDocs),
  EV_COMMAND_AND_ID(CM_SHCOPY, CmCopyOrMove),
  EV_COMMAND_AND_ID(CM_SHMOVE, CmCopyOrMove),
  EV_COMMAND(CM_SHDELETE, CmDelete),
  EV_COMMAND_ENABLE(CM_SHCOPY, CeCopy),
  EV_COMMAND_ENABLE(CM_SHMOVE, CeMove),
  EV_COMMAND_ENABLE(CM_SHDELETE, CeDelete),
  EV_WM_CONTEXTMENU,
  EV_COMMAND(CM_SHCREATESHORTCUT, CmCreateShortcut),
  EV_COMMAND_ENABLE(CM_SHCREATESHORTCUT, CeCreateShortcut),
  EV_COMMAND(CM_SHRENAME, CmRename),
  EV_COMMAND_ENABLE(CM_SHRENAME, CeRename),
//{{TShExWindowRSP_TBL_END}}
END_RESPONSE_TABLE;


//{{TShExWindow Implementation}}


//--------------------------------------------------------
// TShExWindow
// ~~~~~~~~~~
// Construction/Destruction handling.
//
TShExWindow::TShExWindow(TWindow* parent, const char far* title, TModule* module)
:
  TWindow(parent, title, module),
  Item("c:\\")
{
  CoInitialize(NULL); // needed to create shortcuts
}


TShExWindow::~TShExWindow()
{
  Destroy();
}


void TShExWindow::Paint(TDC& dc, bool erase, TRect& rect)
{
  TWindow::Paint(dc, erase, rect);

  try
  {
    // paint the icon
    //
    TIcon icon(Item.GetIcon(TShellItem::Large));
    dc.DrawIcon(kIconXOffset, kIconYOffset, icon);

    // draw the name, path and description of the item
    //
    dc.TextOut(kTextXOffset, kTextYOffset, Item.GetDisplayName());
    dc.TextOut(kTextXOffset, kTextYOffset + kTextHeight,
               Item.GetDisplayName(TShellItem::ForParsing));
    dc.TextOut(kTextXOffset, kTextYOffset + kTextHeight * 2, Item.GetTypeName());
  }
  catch (...) {}
}


void TShExWindow::CmFileOpen()
{
  TShExApp* theApp = TYPESAFE_DOWNCAST(GetApplication(), TShExApp);
  if (!theApp) return;

  // Display standard Open dialog box to select a file name.
  //
  *theApp->FileData.FileName = 0;
  if (TFileOpenDialog(theApp->GetMainWindow(), theApp->FileData).Execute() == IDOK)
  {
    // create the new item
    //
    Item = TShellItem(theApp->FileData.FileName);

    // repaint to show the item
    //
    Invalidate(true);
  }
}


void TShExWindow::CmFileOpenFolder()
{
  try
  {
    // create the shell item for where we start browsing (we will
    // use the desktop). This MUST be a folder.
    //
    TShellItem browseRoot(TShellItem::Desktop);

    // browse for the folder
    //
    TShellItem::TCreateStruct selItem; // the item that is selected by the user
    if (browseRoot.BrowseForFolder(selItem, *this, "Select a folder to open:"))
    {
      // assign the selected item to the current item and repaint
      //
      Item = selItem;
      Invalidate(true);
    }
  }
  catch (TXOwl& x)
  {
    MessageBox(x.why().c_str(), "Error", MB_ICONERROR);
  }
}


void TShExWindow::CmAddToDocs()
{
  Item.AddToRecentDocs();
}


void TShExWindow::CmCopyOrMove(WPARAM id)
{
  try
  {
    // create the shell item for where we start browsing
    // (we will use the desktop)
    //
    TShellItem browseRoot(TShellItem::Desktop);

    // browse for the destination
    //
    TShellItem::TCreateStruct destItem; // the item that is selected by the user
    if (browseRoot.BrowseForFolder(destItem, *this, "Select a destination:"))
    {
      // copy or move to the destination
      //
      switch (id)
      {
        case CM_SHCOPY:
          Item.Copy(destItem, 0, "Copying File", *this);
          break;
        case CM_SHMOVE:
          Item.Move(destItem, 0, "Moving File", *this);
          break;
      }
      
      Invalidate(true);
    }
  }
  catch (TXOwl& x)
  {
    MessageBox(x.why().c_str(), "Error", MB_ICONERROR);
  }
}


void TShExWindow::CmDelete()
{
  try
  {
    TShellItem oldItem = Item;
    Item.Delete(0, "Deleting File", *this);
    Item = oldItem; // in case the user cancelled the delete operation
    Invalidate(true);
  }
  catch (...)
  {
    // TShellItem would be invalid, make it "c:\\" again
    //
    Item = TShellItem("c:\\");
    Invalidate(true);
  }
}


void TShExWindow::CmRename()
{
  char name[MAXPATH];

  // start off with the current name of the item
  //
  strcpy(name, Item.GetDisplayName());

  // ask the user to provide a new name for the item
  //
  if (TInputDialog(this, "Rename File", "Enter the new name:",
                   name, MAXPATH).Execute() == IDOK)
  {
    Item.Rename(name);
    Invalidate(true);
  }
}


void TShExWindow::EvContextMenu(HWND childHwnd, int x, int y)
{
  TWindow::EvContextMenu(childHwnd, x, y);

  // create a popup menu for the item
  //
  TPopupMenu menu;

  // add the shell item's menu commands
  //
  ContextMenu = Item.GetContextMenu(*this);
  ContextMenu->QueryContextMenu(menu, 0, kContextBase, kContextMax, CMF_NORMAL);

  // execute the popup menu
  //
  menu.TrackPopupMenu(TPM_LEFTBUTTON | TPM_RIGHTBUTTON, x, y, 0, *this);
}


TResult TShExWindow::EvCommand(uint id, THandle hWndCtl, uint notifyCode)
{
  TResult result = TWindow::EvCommand(id, hWndCtl, notifyCode);

  if (id >= kContextBase && id <= kContextMax)
  {
    // initialise the command information
    //
    CMINVOKECOMMANDINFO info;
    info.cbSize = sizeof(info);
    info.fMask  = 0;
    info.hwnd   = *this;
    info.lpVerb = (LPCSTR) MAKELONG(id - kContextBase, 0);
    info.lpParameters = NULL;
    info.lpDirectory = NULL;
    info.nShow = SW_SHOW;

    // execute the selected command
    //
    ContextMenu->InvokeCommand(&info);
  }

  return result;
}


void TShExWindow::CmCreateShortcut()
{
  TShExApp* theApp = TYPESAFE_DOWNCAST(GetApplication(), TShExApp);
  if (!theApp) return;

  // Display standard save dialog box to select a file name.
  //
  *theApp->FileData.FileName = 0;
  if (TFileSaveDialog(theApp->GetMainWindow(), theApp->FileData).Execute() == IDOK)
  {
    // get the path and description for the new shortcut
    //
    char path[MAXPATH], drive[MAXDRIVE], dir[MAXDIR], description[MAXPATH], ext[MAXEXT];
    fnsplit(theApp->FileData.FileName, drive, dir, description, ext);
    fnmerge(path, drive, dir, description, ".lnk"); // must specify filename
    strcat(description, ext); // want this to be exactly what the user typed

    // create the shortcut item
    //
    if (!SUCCEEDED(TShellItem::CreateShortCut(Item.GetFullyQualifiedPidl(), path, description)))
      MessageBox("Could not create shortcut.", "Error", MB_ICONERROR);
  }
}


void TShExWindow::CeCopy(TCommandEnabler &tce)
{
  tce.Enable(Item.CanBeCopied());
}


void TShExWindow::CeMove(TCommandEnabler &tce)
{
  tce.Enable(Item.CanBeMoved());
}


void TShExWindow::CeDelete(TCommandEnabler &tce)
{
  tce.Enable(Item.CanBeDeleted());
}


void TShExWindow::CeCreateShortcut(TCommandEnabler &tce)
{
  tce.Enable(Item.CanCreateShortcut());
}


void TShExWindow::CeRename(TCommandEnabler &tce)
{
  tce.Enable(Item.CanBeRenamed());
}

