#include <owl/pch.h>
#include <classlib/arrays.h>

#include <owl/gdiobjec.h>

#include "step3.rh"


// -----------------------------------------------------------
// TShape class
//

class TShape
{
  public:
    virtual void Draw(TDC& dc) = 0;
    virtual void Update(TPoint& p) = 0;
    bool operator==(const TShape& other) { return false; }
};


typedef TIArrayAsVector<TShape> TShapeArray;


// -----------------------------------------------------------
// TLine class
//

class TLine : public TShape
{
  public:
    TLine() {}
    TLine(const TPoint& p1, const TPoint& p2) : P1(p1), P2(p2) {}
    virtual void Draw(TDC& dc);
    virtual void Update(TPoint& p) { P2 = p; }
  protected:
    TPoint P1, P2;
};


void TLine::Draw(TDC& dc)
{
  // Position the graphics "pointer" at the
  // beginning of the line.
  //
  dc.MoveTo(P1);

  // Draw a line from the current position to
  // the other end.
  //
  dc.LineTo(P2);
}


// -----------------------------------------------------------
// TRectangle class
//

class TRectangle : public TShape
{
  public:
    TRectangle() {}
    TRectangle(const TRect& rect) : Rect(rect) {}
    virtual void Draw(TDC& dc);
    virtual void Update(TPoint& p) { Rect.BottomRight() = p; }
  protected:
    TRect Rect;
};


void TRectangle::Draw(TDC& dc)
{
  dc.Rectangle(Rect);
}


// -----------------------------------------------------------
// TEllipse class
//

class TEllipse : public TShape
{
  public:
    TEllipse() {}
    TEllipse(const TRect& rect) : Rect(rect) {}
    virtual void Draw(TDC& dc);
    virtual void Update(TPoint& p) { Rect.BottomRight() = p; }
  protected:
    TRect Rect;
};


void TEllipse::Draw(TDC& dc)
{
  dc.Ellipse(Rect);
}


// -----------------------------------------------------------
// TTutorialClientWindow class
//

class TTutorialClientWindow : public TWindow
{
  public:
    TTutorialClientWindow(TWindow* parent, TModule* module = 0);

  public:
    enum TShapeType {
      Line      = CM_SHAPELINE,
      Rectangle = CM_SHAPERECTANGLE,
      Ellipse   = CM_SHAPEELLIPSE
    };
    static TShapeType ShapeType;

  protected:
    TShapeArray Shapes;
    TShape* DragShape;  // the shape currently being dragged
    TDC* DragDC;        // if this is non-zero, we are dragging

  protected:
    virtual void Paint(TDC& dc, bool erase, TRect& rect);

  protected:
    void EvLButtonDown(uint modKeys, TPoint& point);
    void EvLButtonUp(uint modKeys, TPoint& point);
    void EvMouseMove(uint modKeys, TPoint& point);
    void CmShapeClear();
    DECLARE_RESPONSE_TABLE(TTutorialClientWindow);
};


TTutorialClientWindow::TShapeType TTutorialClientWindow::ShapeType = Line;


DEFINE_RESPONSE_TABLE1(TTutorialClientWindow, TWindow)
  EV_WM_LBUTTONDOWN,
  EV_WM_LBUTTONUP,
  EV_WM_MOUSEMOVE,
  EV_COMMAND(CM_SHAPECLEAR, CmShapeClear),
END_RESPONSE_TABLE;


TTutorialClientWindow::TTutorialClientWindow(TWindow* parent, TModule* module)
: TWindow(parent, 0, module),
  Shapes(1, 0, 1),
  DragDC(0)
{
}


TPen thePen(TColor::LtRed, 5);


void TTutorialClientWindow::EvLButtonDown(uint /*modKeys*/, TPoint& point)
{
  // Begin the drag operation.
  //
  // We select a null brush into the DC because we
  // do not want our drag objects to be filled.
  //
  // SetCapture causes all mouse input to be directed
  // to this window. This ensures that the drag
  // operation is not interrupted.
  //
  DragDC = new TClientDC(*this);
  DragDC->SelectObject((HBRUSH) GetStockObject(NULL_BRUSH));
  SetCapture();

  // Add a new shape to the array of shapes.
  //
  switch (ShapeType)
  {
    case Line:
      DragShape = new TLine(point, point);
      break;
    case Rectangle:
      DragShape = new TRectangle(TRect(point, point));
      break;
    case Ellipse:
      DragShape = new TEllipse(TRect(point, point));
      break;
  }
  Shapes.Add(DragShape);

  // Draw an (inverted) image of the shape.
  //
  int mode = DragDC->SetROP2(R2_NOT);
  DragShape->Draw(*DragDC);
  DragDC->SetROP2(mode);
}


void TTutorialClientWindow::EvLButtonUp(uint /*modKeys*/, TPoint& point)
{
  if (DragDC)
  {
    // Draw an (inverted) image of the shape. (This
    // is to remove any previously drawn image).
    //
    int mode = DragDC->SetROP2(R2_NOT);
    DragShape->Draw(*DragDC);
    DragDC->SetROP2(mode);

    // End the drag operation.
    //
    delete DragDC;
    DragDC = 0;
    ReleaseCapture();

    // Update the shape to its final size.
    //
    DragShape->Update(point);

    // Force the window to be repainted
    //
    Invalidate(true);
  }
}


void TTutorialClientWindow::EvMouseMove(uint /*modKeys*/, TPoint& point)
{
  if (DragDC) // We are in the middle of a drag operation.
  {
    // Draw an (inverted) image of the shape. (This
    // is to remove any previously drawn image).
    //
    int mode = DragDC->SetROP2(R2_NOT);
    DragShape->Draw(*DragDC);
    DragDC->SetROP2(mode);

    // Update the shape to its new size
    //
    DragShape->Update(point);

    // Draw an new (inverted) image of the shape.
    //
    mode = DragDC->SetROP2(R2_NOT);
    DragShape->Draw(*DragDC);
    DragDC->SetROP2(mode);
  }
}


void TTutorialClientWindow::CmShapeClear()
{
  // Remove all of the shapes in the array.
  //
  Shapes.Flush();

  // Force the window to be repainted.
  //
  Invalidate(true);
}


void TTutorialClientWindow::Paint(TDC& dc, bool /*erase*/, TRect& /*rect*/)
{
  //dc.SelectObject(TPen(TColor::LtRed, 5));
  for (int i = 0, iMax = Shapes.GetItemsInContainer(); i < iMax; i++)
    Shapes[i]->Draw(dc);
}


// -----------------------------------------------------------
// TTutorialApp class
//

class TTutorialApp : public TApplication
{
  public:
    TTutorialApp();
    void InitMainWindow();

  protected:
    void CmHelpAbout();
    void CmShapeSelect(uint id);
    void CeShapeSelect(TCommandEnabler& tce);
    DECLARE_RESPONSE_TABLE(TTutorialApp);
};


DEFINE_RESPONSE_TABLE1(TTutorialApp, TApplication)
  EV_COMMAND(CM_HELPABOUT, CmHelpAbout),
  EV_COMMAND_AND_ID(CM_SHAPELINE, CmShapeSelect),
  EV_COMMAND_AND_ID(CM_SHAPERECTANGLE, CmShapeSelect),
  EV_COMMAND_AND_ID(CM_SHAPEELLIPSE, CmShapeSelect),
  EV_COMMAND_ENABLE(CM_SHAPELINE, CeShapeSelect),
  EV_COMMAND_ENABLE(CM_SHAPERECTANGLE, CeShapeSelect),
  EV_COMMAND_ENABLE(CM_SHAPEELLIPSE, CeShapeSelect),
END_RESPONSE_TABLE;


TTutorialApp::TTutorialApp()
: TApplication("Tutorial Application")
{
}


void TTutorialApp::InitMainWindow()
{
  TTutorialClientWindow* client = new TTutorialClientWindow(0);
  TFrameWindow* frame = new TFrameWindow(0, GetName(), client);
  frame->AssignMenu(IDM_MAINMENU);
  frame->Attr.AccelTable = IDM_MAINMENU;
  SetMainWindow(frame);
}


void TTutorialApp::CmHelpAbout()
{
  MessageBox(0, "About", "Help", MB_OK);
}


void TTutorialApp::CmShapeSelect(uint id)
{
  TTutorialClientWindow::ShapeType = (TTutorialClientWindow::TShapeType) id;
}


void TTutorialApp::CeShapeSelect(TCommandEnabler& tce)
{
  tce.SetCheck(TTutorialClientWindow::ShapeType == tce.GetId() ?
               TCommandEnabler::Checked :
               TCommandEnabler::Unchecked);
}


int OwlMain(int /*argc*/, char* /*argv*/[])
{
  TTutorialApp app;
  return app.Run();
}
