#include <owl/pch.h>
#include <owl/opensave.h>
#include <classlib/arrays.h>
#include <classlib/filename.h>

#include "step4.rh"


// -----------------------------------------------------------
// TShape class
//

class TShape : public TStreamableBase
{
  public:
    TShape() {}
    virtual void Draw(TDC& dc) = 0;
    virtual void Update(TPoint& p) = 0;
    bool operator==(const TShape& other) { return false; }

  protected:
    DECLARE_ABSTRACT_STREAMABLE(, TShape, 1);
};


typedef TIArrayAsVector<TShape> TShapeArray;


IMPLEMENT_ABSTRACT_STREAMABLE(TShape);


void* TShape::Streamer::Read(ipstream&, uint32) const
{
  // TShape class contains no data so nothing to read.

  return GetObject();
}


void TShape::Streamer::Write(opstream&) const
{
  // TShape class contains no data so nothing to write.
}


// -----------------------------------------------------------
// TLine class
//

class TLine : public TShape
{
  public:
    TLine() {}
    TLine(const TPoint& p1, const TPoint& p2) : P1(p1), P2(p2) {}
    virtual void Draw(TDC& dc);
    virtual void Update(TPoint& p) { P2 = p; }

  protected:
    TPoint P1, P2;

  protected:
    DECLARE_STREAMABLE(, TLine, 1);
};


IMPLEMENT_STREAMABLE1(TLine, TShape);


void* TLine::Streamer::Read(ipstream& in, uint32) const
{
  // Read in the base object. The cast is important
  // to prevent infinite recursion.
  ReadBaseObject(static_cast<TShape*>(GetObject()), in);

  // Read in the data.
  in >> GetObject()->P1 >> GetObject()->P2;

  return GetObject();
}


void TLine::Streamer::Write(opstream& out) const
{
  // Write out the base object. The cast is important
  // to prevent infinite recursion.
  WriteBaseObject(static_cast<TShape*>(GetObject()), out);

  // Write out the data.
  out << GetObject()->P1 << GetObject()->P2;
}


void TLine::Draw(TDC& dc)
{
  dc.MoveTo(P1);
  dc.LineTo(P2);
}


// -----------------------------------------------------------
// TRectangle class
//

class TRectangle : public TShape
{
  public:
    TRectangle() {}
    TRectangle(const TRect& rect) : Rect(rect) {}
    virtual void Draw(TDC& dc);
    virtual void Update(TPoint& p) { Rect.BottomRight() = p; }

  protected:
    TRect Rect;

  protected:
    DECLARE_STREAMABLE(, TRectangle, 1);
};


IMPLEMENT_STREAMABLE1(TRectangle, TShape);


void* TRectangle::Streamer::Read(ipstream& in, uint32) const
{
  ReadBaseObject(static_cast<TShape*>(GetObject()), in);
  in >> GetObject()->Rect;
  return GetObject();
}


void TRectangle::Streamer::Write(opstream& out) const
{
  WriteBaseObject(static_cast<TShape*>(GetObject()), out);
  out << GetObject()->Rect;
}


void TRectangle::Draw(TDC& dc)
{
  dc.Rectangle(Rect);
}


// -----------------------------------------------------------
// TEllipse class
//

class TEllipse : public TShape
{
  public:
    TEllipse() {}
    TEllipse(const TRect& rect) : Rect(rect) {}
    virtual void Draw(TDC& dc);
    virtual void Update(TPoint& p) { Rect.BottomRight() = p; }

  protected:
    TRect Rect;

  protected:
    DECLARE_STREAMABLE(, TEllipse, 1);
};


IMPLEMENT_STREAMABLE1(TEllipse, TShape);


void* TEllipse::Streamer::Read(ipstream& in, uint32) const
{
  ReadBaseObject(static_cast<TShape*>(GetObject()), in);
  in >> GetObject()->Rect;
  return GetObject();
}


void TEllipse::Streamer::Write(opstream& out) const
{
  WriteBaseObject(static_cast<TShape*>(GetObject()), out);
  out << GetObject()->Rect;
}


void TEllipse::Draw(TDC& dc)
{
  dc.Ellipse(Rect);
}


// -----------------------------------------------------------
// TTutorialClientWindow class
//

class TTutorialClientWindow : public TWindow
{
  public:
    TTutorialClientWindow(TWindow* parent, TModule* module = 0);

  public:
    enum TShapeType {
      Line      = CM_SHAPELINE,
      Rectangle = CM_SHAPERECTANGLE,
      Ellipse   = CM_SHAPEELLIPSE
    };
    static TShapeType ShapeType;

  protected:
    TShapeArray Shapes;
    TShape* DragShape;
    TDC* DragDC;
    string FileName;
    bool Dirty;

  public:
    void New();
    void Open(const char* filename);
    void Save(const char* filename);
    bool DoSaveAs();

  public:
    virtual void Paint(TDC& dc, bool erase, TRect& rect);
    virtual bool CanClose();

  protected:
    void EvLButtonDown(uint modKeys, TPoint& point);
    void EvLButtonUp(uint modKeys, TPoint& point);
    void EvMouseMove(uint modKeys, TPoint& point);
    void CmShapeClear();
    void CmFileSave();
    void CeFileSave(TCommandEnabler& tce);
    void CmFileSaveAs();
    DECLARE_RESPONSE_TABLE(TTutorialClientWindow);
};


TTutorialClientWindow::TShapeType TTutorialClientWindow::ShapeType = Line;


DEFINE_RESPONSE_TABLE1(TTutorialClientWindow, TWindow)
  EV_WM_LBUTTONDOWN,
  EV_WM_LBUTTONUP,
  EV_WM_MOUSEMOVE,
  EV_COMMAND(CM_SHAPECLEAR, CmShapeClear),
  EV_COMMAND(CM_FILESAVE, CmFileSave),
  EV_COMMAND(CM_FILESAVEAS, CmFileSaveAs),
  EV_COMMAND_ENABLE(CM_FILESAVE, CeFileSave),
END_RESPONSE_TABLE;


TTutorialClientWindow::TTutorialClientWindow(TWindow* parent, TModule* module)
: TWindow(parent, 0, module),
  Shapes(1, 0, 1),
  DragDC(0),
  Dirty(false)
{
}


void TTutorialClientWindow::EvLButtonDown(uint /*modKeys*/, TPoint& point)
{
  // Begin the drag operation.
  //
  DragDC = new TWindowDC(*this);
  DragDC->SelectObject((HBRUSH) GetStockObject(NULL_BRUSH));
  SetCapture();

  // Add a new shape to the array of shapes.
  //
  switch (ShapeType)
  {
    case Line:
      DragShape = new TLine(point, point);
      break;
    case Rectangle:
      DragShape = new TRectangle(TRect(point, point));
      break;
    case Ellipse:
      DragShape = new TEllipse(TRect(point, point));
      break;
  }
  Shapes.Add(DragShape);

  // Set the dirty flag to indicate that the
  // document has changed.
  //
  Dirty = true;

  // Draw an (inverted) image of the shape.
  //
  int mode = DragDC->SetROP2(R2_NOT);
  DragShape->Draw(*DragDC);
  DragDC->SetROP2(mode);
}


void TTutorialClientWindow::EvLButtonUp(uint /*modKeys*/, TPoint& point)
{
  if (DragDC)
  {
    // Draw an (inverted) image of the shape. (This
    // is to remove any previously drawn image).
    //
    int mode = DragDC->SetROP2(R2_NOT);
    DragShape->Draw(*DragDC);
    DragDC->SetROP2(mode);

    // End the drag operation.
    //
    delete DragDC;
    DragDC = 0;
    ReleaseCapture();

    // Update the shape to its final size.
    //
    DragShape->Update(point);

    // Force the window to be repainted
    //
    Invalidate(true);
  }
}


void TTutorialClientWindow::EvMouseMove(uint /*modKeys*/, TPoint& point)
{
  if (DragDC) // We are in the middle of a drag operation.
  {
    // Draw an (inverted) image of the shape. (This
    // is to remove any previously drawn image).
    //
    int mode = DragDC->SetROP2(R2_NOT);
    DragShape->Draw(*DragDC);
    DragDC->SetROP2(mode);

    // Update the shape to its new size
    //
    DragShape->Update(point);

    // Draw an new (inverted) image of the shape.
    //
    mode = DragDC->SetROP2(R2_NOT);
    DragShape->Draw(*DragDC);
    DragDC->SetROP2(mode);
  }
}


void TTutorialClientWindow::CmShapeClear()
{
  Shapes.Flush();
  Invalidate(true);
  Dirty = true;
}


void TTutorialClientWindow::Paint(TDC& dc, bool /*erase*/, TRect& /*rect*/)
{
  for (int i = 0, iMax = Shapes.GetItemsInContainer(); i < iMax; i++)
    Shapes[i]->Draw(dc);
}


bool TTutorialClientWindow::CanClose()
{
  if (Dirty) // there is something to save
  {
    // Display a message box asking the user whether
    // they wish to save the changes.
    //
    string untitled = "Untitled";
    string msg = "The " + (FileName.length() ? FileName : untitled) +
                 " file has changed.\n\nDo you wish to save the changes?";
    int answer = MessageBox(msg.c_str(), GetApplication()->GetName(),
                            MB_YESNOCANCEL | MB_ICONEXCLAMATION);

    // Depending on what the user picked, save,
    // don't save, or don't close.
    //
    switch (answer)
    {
      case IDYES:
        // If the document is not untitled, just save it.
        //
        if (FileName.length())
        {
          Save(FileName.c_str());
          return true;
        }

        // Document is untitled, so show save dialog.
        //
        if (DoSaveAs())
          return true;

        // User cancelled save dialog, so don't close.
        //
        return false;

      case IDNO:
        return true;

      case IDCANCEL:
        return false;
    }
  }

  return true;
}



void TTutorialClientWindow::New()
{
  // Remove all the current shapes.
  //
  Shapes.Flush();

  // Force the window to be repainted.
  Invalidate(true);

  // Clear the stored filename.
  //
  FileName = "";

  // Clear the dirty flag to indicate that
  // no changes have been made.
  Dirty = false;

  // Update the frame window caption.
  //
  Parent->SetCaption((string("Untitled - ") + GetApplication()->GetName()).c_str());
}


void TTutorialClientWindow::Open(const char* filename)
{
  // Remove all the current shapes.
  //
  Shapes.Flush();

  // Open the file as a persistent stream.
  //
  ifpstream in(filename);
  if (in.bad())
  {
    MessageBox("Unable to open file for reading.", filename, MB_ICONERROR);
    return;
  }

  // Read each of the shapes from the file.
  // Reading each shape as a pointer lets us
  // read in a TShape where the actual object
  // pointed to is one of the derived classes.
  //
  int numShapes;
  TShape* shape;
  in >> numShapes;
  while (numShapes--)
  {
    in >> shape;
    Shapes.Add(shape);
  }

  // Force the window to be repainted.
  //
  Invalidate(true);

  // Remember the filename for later use.
  //
  FileName = filename;

  // Clear the dirty flag to indicate that
  // no changes have been made.
  Dirty = false;

  // Update the frame window caption.
  //
  string caption = TFileName(FileName).Title() + " - " +
                   GetApplication()->GetName();
  Parent->SetCaption(caption.c_str());
}


void TTutorialClientWindow::Save(const char* filename)
{
  // Open the file as a persistent stream.
  //
  ofpstream out(filename);
  if (out.bad())
  {
    MessageBox("Unable to open file for writing.", filename, MB_ICONERROR);
    return;
  }

  // Write each of the shapes out to the file. By
  // writing the shape out using a pointer to it
  // (rather than the object itself or a reference)
  // the persistent stream allows us to store and
  // read objects without needing to know their
  // exact class (in this case, one of TLine,
  // TRectange or TEllipse).
  //
  int numShapes = Shapes.GetItemsInContainer();
  out << numShapes;
  for (int i = 0; i < numShapes; i++)
    out << Shapes[i];

  // Clear the dirty flag to indicate that
  // no changes have been made.
  Dirty = false;
}


void TTutorialClientWindow::CmFileSave()
{
  if (!FileName.length())
    CmFileSaveAs();
  else Save(FileName.c_str());
}


void TTutorialClientWindow::CeFileSave(TCommandEnabler& tce)
{
  // Don't want to enable the save menu item
  // unless the document is dirty.
  //
  tce.Enable(Dirty);
}


void TTutorialClientWindow::CmFileSaveAs()
{
  DoSaveAs();
}


bool TTutorialClientWindow::DoSaveAs()
{
  // Create a data object containing information for
  // the save file dialog.
  //
  TOpenSaveDialog::TData fileData;
  fileData.Flags = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT | OFN_EXPLORER;
  fileData.SetFilter("Shape Files (*.shp)|*.shp|All Files (*.*)|*.*|");
  fileData.DefExt = "shp";
  strcpy(fileData.FileName, FileName.c_str());

  // Display the File Save dialog, and if the user closes
  // it by pressing OK, save the file.
  //
  if (TFileSaveDialog(GetApplication()->GetMainWindow(), fileData).Execute() == IDOK)
  {
    // Save the shapes to the file.
    //
    Save(fileData.FileName);

    // Remember the filename for later.
    //
    FileName = fileData.FileName;

    // Update the frame window caption.
    //
    string caption = TFileName(FileName).Title() + " - " +
                     GetApplication()->GetName();
    Parent->SetCaption(caption.c_str());

    // Return true to indicate that the file was saved.
    //
    return true;
  }

  return false;
}


// -----------------------------------------------------------
// TTutorialApp class
//

class TTutorialApp : public TApplication
{
  public:
    TTutorialApp();
    void InitMainWindow();

  protected:
    void CmHelpAbout();
    void CmShapeSelect(uint id);
    void CeShapeSelect(TCommandEnabler& tce);
    void CmFileNew();
    void CmFileOpen();
    DECLARE_RESPONSE_TABLE(TTutorialApp);
};


DEFINE_RESPONSE_TABLE1(TTutorialApp, TApplication)
  EV_COMMAND(CM_HELPABOUT, CmHelpAbout),
  EV_COMMAND_AND_ID(CM_SHAPELINE, CmShapeSelect),
  EV_COMMAND_AND_ID(CM_SHAPERECTANGLE, CmShapeSelect),
  EV_COMMAND_AND_ID(CM_SHAPEELLIPSE, CmShapeSelect),
  EV_COMMAND_ENABLE(CM_SHAPELINE, CeShapeSelect),
  EV_COMMAND_ENABLE(CM_SHAPERECTANGLE, CeShapeSelect),
  EV_COMMAND_ENABLE(CM_SHAPEELLIPSE, CeShapeSelect),
  EV_COMMAND(CM_FILENEW, CmFileNew),
  EV_COMMAND(CM_FILEOPEN, CmFileOpen),
END_RESPONSE_TABLE;


TTutorialApp::TTutorialApp()
: TApplication("Tutorial Application")
{
}


void TTutorialApp::InitMainWindow()
{
  TTutorialClientWindow* client = new TTutorialClientWindow(0);
  TFrameWindow* frame = new TFrameWindow(0, GetName(), client);
  frame->AssignMenu(IDM_MAINMENU);
  frame->Attr.AccelTable = IDM_MAINMENU;
  SetMainWindow(frame);
  client->New();
}


class TWaitCursor
{
public:
  TWaitCursor(TModule* module = 0, TResId resId = IDC_WAIT)
  : OldCursor(0)
  {
    HCURSOR cursor = LoadCursor(module ? (HINSTANCE) *module : 0, resId);
    if (cursor) OldCursor = SetCursor(cursor);
  }

  ~TWaitCursor()
  {
    if (OldCursor) SetCursor(OldCursor);
  }

private:
  HCURSOR OldCursor;
};


void TTutorialApp::CmHelpAbout()
{
  TWaitCursor wait;

  Sleep(2000);

  //MessageBox(0, "About", "Help", MB_OK);
}


void TTutorialApp::CmShapeSelect(uint id)
{
  TTutorialClientWindow::ShapeType = (TTutorialClientWindow::TShapeType) id;
}


void TTutorialApp::CeShapeSelect(TCommandEnabler& tce)
{
  tce.SetCheck(TTutorialClientWindow::ShapeType == tce.GetId() ?
               TCommandEnabler::Checked :
               TCommandEnabler::Unchecked);
}


void TTutorialApp::CmFileNew()
{
  // Get a pointer to the client window.
  //
  TTutorialClientWindow* client;
  client = dynamic_cast<TTutorialClientWindow*>(GetMainWindow()->GetClientWindow());

  // Check if the user wants to save the current image.
  //
  if (!client->CanClose()) return;

  // Tell the client window to "new" itself.
  //
  if (client) client->New();
}


void TTutorialApp::CmFileOpen()
{
  // Get a pointer to the client window.
  //
  TTutorialClientWindow* client;
  client = dynamic_cast<TTutorialClientWindow*>(GetMainWindow()->GetClientWindow());
  if (!client) return;

  // Check if the user wants to save the current image.
  //
  if (!client->CanClose()) return;

  // Create a data object containing information for
  // the open file dialog.
  //
  TOpenSaveDialog::TData fileData;
  fileData.Flags = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT | OFN_EXPLORER;
  fileData.SetFilter("Shape Files (*.shp)|*.shp|All Files (*.*)|*.*|");
  fileData.DefExt = "shp";

  // Display the File Open dialog, and if the user closes
  // it by pressing OK, open the selected file.
  //
  if (TFileOpenDialog(GetMainWindow(), fileData).Execute() == IDOK)
  {
    // Tell the client window to open the file.
    //
    if (client) client->Open(fileData.FileName);
  }
}


int OwlMain(int /*argc*/, char* /*argv*/[])
{
  TTutorialApp app;
  return app.Run();
}